@@ -1,7 +1,7 @@
 /*
  * mdadm - manage Linux "md" devices aka RAID arrays.
  *
- * Copyright (C) 2001-2006 Neil Brown <neilb@suse.de>
+ * Copyright (C) 2001-2009 Neil Brown <neilb@suse.de>
  *
  *
  *    This program is free software; you can redistribute it and/or modify
@@ -19,12 +19,7 @@
  *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  *
  *    Author: Neil Brown
- *    Email: <neilb@cse.unsw.edu.au>
- *    Paper: Neil Brown
- *           School of Computer Science and Engineering
- *           The University of New South Wales
- *           Sydney, 2052
- *           Australia
+ *    Email: <neilb@suse.de>
  */
 #include	"mdadm.h"
 #include	"dlink.h"
@@ -69,7 +64,7 @@ int Grow_Add_device(char *devname, int fd, char *newdev)
 		return 1;
 	}
 
-	nfd = open(newdev, O_RDWR|O_EXCL);
+	nfd = open(newdev, O_RDWR|O_EXCL|O_DIRECT);
 	if (nfd < 0) {
 		fprintf(stderr, Name ": cannot open %s\n", newdev);
 		return 1;
@@ -396,7 +391,8 @@ struct mdp_backup_super {
 	__u64	arraystart;
 	__u64	length;
 	__u32	sb_csum;	/* csum of preceeding bytes. */
-};
+	__u8 pad[512-68];
+} __attribute__((aligned(512))) bsb;
 
 int bsb_csum(char *buf, int len)
 {
@@ -420,7 +416,6 @@ int Grow_reshape(char *devname, int fd, int quiet, char *backup_file,
 	struct mdu_array_info_s array;
 	char *c;
 
-	struct mdp_backup_super bsb;
 	struct supertype *st;
 
 	int nlevel, olevel;
@@ -720,7 +715,8 @@ int Grow_reshape(char *devname, int fd, int quiet, char *backup_file,
 		 * a leading superblock 4K earlier.
 		 */
 		for (i=array.raid_disks; i<d; i++) {
-			char buf[4096];
+			char abuf[4096+512];
+			char *buf = (char*)(((unsigned long)abuf+511)& ~511);
 			if (i==d-1 && backup_file) {
 				/* This is the backup file */
 				offsets[i] = 8;
@@ -731,7 +727,7 @@ int Grow_reshape(char *devname, int fd, int quiet, char *backup_file,
 				fprintf(stderr, Name ": could not seek...\n");
 				goto abort;
 			}
-			memset(buf, 0, sizeof(buf));
+			memset(buf, 0, 4096);
 			bsb.devstart = __cpu_to_le64(offsets[i]);
 			bsb.sb_csum = bsb_csum((char*)&bsb, ((char*)&bsb.sb_csum)-((char*)&bsb));
 			memcpy(buf, &bsb, sizeof(bsb));
@@ -793,7 +789,7 @@ int Grow_reshape(char *devname, int fd, int quiet, char *backup_file,
 			if (lseek64(fdlist[i], (offsets[i]+last_block)<<9, 0) < 0 ||
 			    write(fdlist[i], &bsb, sizeof(bsb)) != sizeof(bsb) ||
 			    fsync(fdlist[i]) != 0) {
-				fprintf(stderr, Name ": %s: fail to save metadata for critical region backups.\n",
+				fprintf(stderr, Name ": %s: failed to save metadata for critical region backups.\n",
 					devname);
 				goto abort_resume;
 			}
@@ -808,12 +804,21 @@ int Grow_reshape(char *devname, int fd, int quiet, char *backup_file,
 		/* wait for reshape to pass the critical region */
 		while(1) {
 			unsigned long long comp;
+
 			if (sysfs_get_ll(sra, NULL, "sync_completed", &comp)<0) {
 				sleep(5);
 				break;
 			}
 			if (comp >= nstripe)
 				break;
+			if (comp == 0) {
+				/* Maybe it finished already */
+				char action[20];
+				if (sysfs_get_str(sra, NULL, "sync_action",
+						  action, 20) > 0 &&
+				    strncmp(action, "reshape", 7) != 0)
+					break;
+			}
 			sleep(1);
 		}
 
@@ -882,7 +887,6 @@ int Grow_restart(struct supertype *st, struct mdinfo *info, int *fdlist, int cnt
 
 	for (i=old_disks-(backup_file?1:0); i<cnt; i++) {
 		struct mdinfo dinfo;
-		struct mdp_backup_super bsb;
 		char buf[4096];
 		int fd;
 
